//////////////////////////////////////////////////////////////////////////
// HugeInt.cpp
//////////////////////////////////////////////////////////////////////////
// Implementation of the Huge Integer class. This is a data type that
// supports unlimited sized numeric values. Think of it as an integer that
// can be ANY value, no matter how large.
// All standard math operators are applicable (+ - * / % ++ --)
// All logical operators apply (> < >= <= == !=)
// Conversion to HINT is possible from the following types:
//   int, double, const char*
// Standard insertion and extraction (<< and >>) operators also apply.
// Use HINT as you would a normal int variable. The difference is that
// you can assign it a string (if you want to assign a huge number that's
// larger than the largest built-in variable). For example:
//   HINT huge = "999000123000999888777666555444333222111987654321888666"
//   cout << huge * "99999999999999999" << endl;
// If you cast or assign a double precision value, any decimal value will
// be truncated and not rounded.
//////////////////////////////////////////////////////////////////////////
// This is free source. It may be a bit faulty. I haven't tested it to
// any real extremes. If you expand this source, please send me your
// version. I'd appreciate it.
//////////////////////////////////////////////////////////////////////////
// 08/08/02 - ROMMEL SANTOR - lucky760@yahoo.com
//////////////////////////////////////////////////////////////////////////

#include "StdAfx.h"//in case you don't know, this is only needed for MSVC AppWizard apps

#include "HugeInt.h"
    
//////////////////////////////////////////////////////////////////////////
// Friends
//////////////////////////////////////////////////////////////////////////
std::ostream& operator<<(std::ostream &out,const HINT &rhs) {
  if (rhs.m_bNeg) 
    out << '-';
  for (int i=rhs.m_Digits.GetSize()-1;i>=0;--i)
    out << (int)(rhs.m_Digits[i]);
  return out;
}

std::istream& operator>>(std::istream &in,HINT &rhs) {  
  GrowArray<char> grow;

  int nPos = 0;

  char ch;
  in.get(ch);

  while (((isspace(ch) && nPos == 0) || !isspace(ch)) && in.good()) {    
    if (isspace(ch) && nPos == 0) {
      in.get(ch);
      continue;
    }

    grow[nPos++] = ch;
    grow[nPos] = 0;

    in.get(ch);
  }            

  if (nPos > 0)
    rhs = grow.GetData();

  return in;
}

const HINT operator-(int nVal,const HINT &rhs) { return (HINT)nVal - rhs; }
const HINT operator+(int nVal,const HINT &rhs) { return rhs + nVal; }
const HINT operator*(int nVal,const HINT &rhs) { return rhs * nVal; }
const HINT operator/(int nVal,const HINT &rhs) { return (HINT)nVal / rhs; }
const HINT operator%(int nVal,const HINT &rhs) { return (HINT)nVal % rhs; }
  
const HINT operator-(const char *szVal,const HINT &rhs) { return (HINT)szVal - rhs; }
const HINT operator+(const char *szVal,const HINT &rhs) { return (HINT)szVal + rhs; }
const HINT operator*(const char *szVal,const HINT &rhs) { return (HINT)szVal * rhs; }
const HINT operator/(const char *szVal,const HINT &rhs) { return (HINT)szVal / rhs; }
const HINT operator%(const char *szVal,const HINT &rhs) { return (HINT)szVal % rhs; }
  
//////////////////////////////////////////////////////////////////////////
// Miscellaneous Member Functions
//////////////////////////////////////////////////////////////////////////
void HINT::RemoveLeadingZeros() {
  if (*this == 0)
    m_Digits.Resize(0);

  while (m_Digits.GetSize() && m_Digits[m_Digits.GetSize()-1] == 0)
    m_Digits.Resize(m_Digits.GetSize() - 1);
}

//////////////////////////////////////////////////////////////////////////
// Assignment operators
//////////////////////////////////////////////////////////////////////////
HINT& HINT::operator=(const char *szVal) {
  int nLen = strlen(szVal);
  const char *pNeg = strstr(szVal,"-");

  m_bNeg = pNeg != NULL;

  int i=0;
  if (m_bNeg) {
    assert ((pNeg == szVal) || (pNeg == szVal+nLen-1));//'-' must be at head or tail  

    if (pNeg == szVal)
      i = 1;
    else
      --nLen;
  }

  bool bZero = true;

  Alloc(0);
  Alloc(nLen);

  for ( ;i<nLen;++i) {    
    if (szVal[i] != '-') {
      if (szVal[i] == '.')
        break;
      assert(szVal[i] >= '0' && szVal[i] <= '9');//numbers only!
      m_Digits[nLen-i-1] = szVal[i] - '0';

      if (szVal[i] != '0')
        bZero = false;
    }
    else if (m_bNeg)
      assert(pNeg == (szVal+i));//must be one and only one '-'
  }

  if (bZero)
    m_bNeg = false;
  else
    RemoveLeadingZeros();

  return *this;  
}
  
HINT& HINT::operator=(int nVal) {
  char szTemp[/*CHARR::*/DEF_SIZE];
  itoa(nVal,szTemp,10);
  return operator=(szTemp);
}
  
HINT& HINT::operator=(const HINT &rhs) {
  m_bNeg = rhs.m_bNeg;
  m_Digits = rhs.m_Digits;
  return *this;
}

HINT& HINT::operator=(double dVal) {
  int  nDec, nSign;
  
  char *pBuf = _fcvt(dVal,255,&nDec,&nSign);
  
  GrowArray<char> grow;
  for (int i=0;i<nDec;++i) {
    grow[i] = pBuf[i];
    grow[i+1] = 0;
  }
  
  operator=(grow.GetData());
  m_bNeg = dVal < 0;
  
  return *this;
}

//////////////////////////////////////////////////////////////////////////
// Logical operators
//////////////////////////////////////////////////////////////////////////
bool HINT::operator>(const HINT &rhs) const {
  if (m_bNeg && !rhs.m_bNeg)
    return false;
  else if (!m_bNeg && rhs.m_bNeg)
    return true;

  if (rhs == 0 && *this != 0)
    return !m_bNeg;
  else if (rhs != 0 && *this == 0)
    return rhs.m_bNeg;

  int nSize = m_Digits.GetSize();
  int nNewSize = rhs.m_Digits.GetSize();
    
  bool bRet = false;
  if (nSize > nNewSize)
    bRet = true;
  else if (nSize < nNewSize)
    bRet = false;
  else {
    for (int i=nSize-1;i>=0;--i) {
      if(m_Digits[i] > rhs.m_Digits[i]) {
        bRet = true;
        break;
      }      
      else if (m_Digits[i] < rhs.m_Digits[i])
        break;
    }
  }
  return m_bNeg ? !bRet : bRet;
}
  
bool HINT::operator==(const HINT &rhs) const {
  bool bZero = true;
  int i;//MSVC is very retarded

  for (i=0;i<m_Digits.GetSize() && bZero;++i)
    if (m_Digits[i] != 0)
      bZero = false;

  bool bRhsZero = true;

  for (i=0;i<rhs.m_Digits.GetSize() && bRhsZero;++i)
    if (rhs.m_Digits[i] != 0)
      bRhsZero = false;

  if ((bZero && !bRhsZero) || (!bZero && bRhsZero))
    return false;

  if ((m_bNeg && !rhs.m_bNeg) || (!m_bNeg && rhs.m_bNeg))
    return false;

  int nSize = m_Digits.GetSize();
  int nNewSize = rhs.m_Digits.GetSize();
    
  if (nSize != nNewSize)
    return false;
  else
    for (int i=nSize-1;i>=0;--i)
      if(m_Digits[i] != rhs.m_Digits[i])
        return false;

  return true;
}

//////////////////////////////////////////////////////////////////////////
// Conversion
//////////////////////////////////////////////////////////////////////////
char* HINT::HToStr(char *szDest,int nMaxLen) {//nMaxLen does not include the NULL
  if (nMaxLen == 0)
    return szDest;

  int index = 0;

  if (m_bNeg) 
    szDest[index++] = '-';

  for (int i=m_Digits.GetSize()-1;i >= 0 && index < nMaxLen;--i)
    szDest[index++] = static_cast<char>(m_Digits[i] + '0');

  szDest[index] = 0;

  return szDest;
}

//////////////////////////////////////////////////////////////////////////
// Mathematical Operators
//////////////////////////////////////////////////////////////////////////

const HINT HINT::operator-() { return operator*(-1); }
  
HINT& HINT::operator++() { return operator+=(1); }//prefix
HINT& HINT::operator--() { return operator-=(1); }//prefix
const HINT HINT::operator++(int) { HINT temp=*this; operator+=(1); return temp; }//postfix
const HINT HINT::operator--(int) { HINT temp=*this; operator-=(1); return temp; }//postfix 

// -----------------------------------------------------------------------
// Addition operator
// -----------------------------------------------------------------------
const HINT HINT::operator+(const HINT &rhs) const {
  if (m_bNeg && !rhs.m_bNeg) {
    HINT a = *this;
    a.m_bNeg = false;
    return rhs - a;
  }
  else if (!m_bNeg && rhs.m_bNeg) {
    HINT a = rhs;
    a.m_bNeg = false;
    return *this - a;
  }

  if (*this == 0)
    return rhs;
  else if (rhs == 0)
    return *this;
    
  HINT temp; //the longer number
  const HINT *other;//the other number
  if (m_Digits.GetSize() >= rhs.m_Digits.GetSize()) {
    temp = *this;
    other = &rhs;
  }
  else {
    temp = rhs;
    other = this;
  }
  temp.m_Digits.Resize(temp.m_Digits.GetSize() + 1);//temp.m_Digits.GetSize() += 1;
  int i;
  /* do the addition */
  for (i=0;i<other->m_Digits.GetSize();++i) {
    temp.m_Digits[i] += other->m_Digits[i];
    if (temp.m_Digits[i] > 9) {
      temp.m_Digits[i] -= 10;
      temp.m_Digits[i+1] += 1;
    }
  }
  /* take care of remaining carries */
  while ((temp.m_Digits[i] > 9) && (i < temp.m_Digits.GetSize())) {
    temp.m_Digits[i] -= 10;
    temp.m_Digits[++i] += 1;
  }
  /* remove leading zeros, if any */
  temp.RemoveLeadingZeros();
  return temp;
}

// -----------------------------------------------------------------------
// Subtraction operator
// -----------------------------------------------------------------------
const HINT HINT::operator-(const HINT &rhs) const {
  HINT temp = *this;

  if (temp == 0)
    return rhs * -1;
  else if (rhs == 0)
    return temp;
    
  if (temp.m_bNeg) {
    //return (temp*-1) + (rhs.m_bNeg ? rhs*-1 : rhs);
    HINT a = temp;
    HINT b = rhs;
    a.m_bNeg = false;
    b.m_bNeg = false;

    if (rhs.m_bNeg)// -1 - (-2)
      temp = b - a;      
    else// -1 - 5
      temp = a + b;

    temp.m_bNeg = true;
    return temp;
  }
  else if (temp == rhs)
    return 0;
  else if (temp < rhs) {
    temp = rhs - temp;
    temp.m_bNeg = true;
    return temp;
  }  

  int i=0;
  for( ;i<rhs.m_Digits.GetSize();++i) {
    if (temp.m_Digits[i] < 0) {
      temp.m_Digits[i] += 10;
      if ((i+1) < temp.m_Digits.GetSize())
        temp.m_Digits[i+1] -= 1;
    }
    if (temp.m_Digits[i] < rhs.m_Digits[i]) {
      temp.m_Digits[i] += 10;
      if ((i+1) < temp.m_Digits.GetSize())
        temp.m_Digits[i+1] -= 1;
    }
    temp.m_Digits[i] -= rhs.m_Digits[i];
  }

  for ( ;i<temp.m_Digits.GetSize();++i) {
    if (temp.m_Digits[i] < 0) {
      temp.m_Digits[i] += 10;
      temp.m_Digits[i+1] -= 1;
    }
    else
      break;
  }

  // remove leading zeros, if any
  temp.RemoveLeadingZeros();

  return temp;
}

// -----------------------------------------------------------------------
// Multiplication operator
// -----------------------------------------------------------------------
const HINT HINT::operator*(const HINT &rhs) const {   
  HINT temp = *this;
  
  if (temp == 0 || rhs == 0)
    return 0;
    
  if (rhs > temp)
    return rhs * temp;
  
  HINT product = 0;
  bool bNeg = ((!m_bNeg && rhs.m_bNeg) || (m_bNeg && !rhs.m_bNeg));
  
  int i;
  // do the multiply
  int nOffset = 0;
  for (i=0;i<temp.m_Digits.GetSize();++i) {
    HINT subTotal;
    subTotal.Alloc(temp.m_Digits.GetSize() + rhs.m_Digits.GetSize());

    for (int j=0;j<rhs.m_Digits.GetSize();++j) {
      int nCarry = 0;
      int nProd = temp.m_Digits[i] * rhs.m_Digits[j];

      if (nProd > 9) {
        nCarry += nProd / 10;
        nProd -= nCarry * 10;
      }

      subTotal.m_Digits[j+nOffset] += nProd;
      subTotal.m_Digits[j+nOffset+1] += nCarry;

      if (subTotal.m_Digits[j+nOffset] > 9) {
        subTotal.m_Digits[j+nOffset] -= 10;
        subTotal.m_Digits[j+nOffset+1] += 1;
      }
    }

    if (subTotal != 0) {
      subTotal.RemoveLeadingZeros();
      product += subTotal; 
    }

    ++nOffset;
  }

  // remove leading zeros, if any
  product.RemoveLeadingZeros();
  
  // restore sign
  product.m_bNeg = bNeg;

  return product;
}

// -----------------------------------------------------------------------
// Division operator
// -----------------------------------------------------------------------
const HINT HINT::operator/(const HINT &denom) const {
  assert(denom != 0);
  
  HINT numer = *this;

  if (numer < denom)
    return 0;
  
  bool bNeg = ((!m_bNeg && denom.m_bNeg) || (m_bNeg && !denom.m_bNeg));
  
  HINT nextNumer = numer;
  int nNumerPos = numer.m_Digits.GetSize()-1;
  
  char *pQuotient = new char[numer.m_Digits.GetSize()+1];
  *pQuotient = 0;
  int nQPos = 0;

  bool bFirstTime = true;
  
  while /*(nextNumer >= denom) {*/(nNumerPos >= 0) {
    int nLen = nextNumer.m_Digits.GetSize();   
      
    int nNumerLen = 0;  
    char *pNumer = new char[nLen+1];
    *pNumer = 0;
    strcpy(pNumer,"0");

    --nLen;

    while (HINT(pNumer) < denom && nLen >= 0) {
      if (bFirstTime)
        --nNumerPos;
      pNumer[nNumerLen++] = nextNumer.m_Digits[nLen--] + '0';
      pNumer[nNumerLen] = 0;
    }

    if (!bFirstTime)
      --nNumerPos;
    
    bFirstTime = false;
    
    int nNextDigit = nNumerPos >= 0  ?  numer.m_Digits[nNumerPos]  :  -1;    
  
    int nMax=0;
    for (int i=1;i<10;++i) {    
      if ((denom*i) > HINT(pNumer))
        break;    
      nMax = i;
    }
  
    pQuotient[nQPos++] = nMax + '0';
    pQuotient[nQPos] = 0;

    nextNumer = HINT(pNumer)-(denom*nMax);
    if (nNextDigit >= 0) {
      nextNumer *= 10;
      nextNumer += nNextDigit;
    }
  
    delete [] pNumer;
  }
  
  HINT quotient = pQuotient;
  delete [] pQuotient;
  
  quotient.m_bNeg = bNeg;
      
  return quotient;
}

// -----------------------------------------------------------------------
// Modulus operator
// -----------------------------------------------------------------------
const HINT HINT::operator%(const HINT &denom) const {
  assert(denom != 0);
  
  HINT numer = *this;

  if (numer < denom)
    return numer;
  
  HINT nextNumer = numer;
  int nNumerPos = numer.m_Digits.GetSize()-1;

  HINT remainder;
  
  bool bFirstTime = true;
  
  while (nNumerPos >= 0) {
    int nLen = nextNumer.m_Digits.GetSize();   
      
    int nNumerLen = 0;  
    char *pNumer = new char[nLen+1];
    *pNumer = 0;
    strcpy(pNumer,"0");

    --nLen;

    while (HINT(pNumer) < denom && nLen >= 0) {
      if (bFirstTime)
        --nNumerPos;
      pNumer[nNumerLen++] = nextNumer.m_Digits[nLen--] + '0';
      pNumer[nNumerLen] = 0;
    }

    if (!bFirstTime)
      --nNumerPos;

    bFirstTime = false;

    int nNextDigit = nNumerPos >= 0  ?  numer.m_Digits[nNumerPos]  :  -1;    

    int nMax=0;
    for (int i=1;i<10;++i) {    
      if ((denom*i) > HINT(pNumer))
        break;    
      nMax = i;
    }

    nextNumer = HINT(pNumer)-(denom*nMax);
    if (nNextDigit >= 0) {
      nextNumer *= 10;
      nextNumer += nNextDigit;
    }  

    delete [] pNumer;

    remainder = nextNumer;      
    remainder.m_bNeg = remainder != 0 ? numer.m_bNeg : false;
  }
     
  return remainder;
}


//////////////////////////////////////////////////////////////////////////
// FREE FUNCTIONS
//////////////////////////////////////////////////////////////////////////
std::istream& getline(std::istream &in,HINT &rhs,char delim) {
  GrowArray<char> grow;
  int nIndex = 0;
  while (in.peek() != delim) {
    grow[nIndex++] = in.get();
    grow[nIndex] = 0;
  }
  rhs = grow.GetData();
  return in;
}
